import pytest

from flake8_warnings._extractors import WarningsExtractor

from .helpers import e, p


def test_module_deprecated():
    r = e(WarningsExtractor, p("""
        import warnings
        from astroid.nodes.node_classes import (
            AsyncFor,
        )
        warnings.warn(
            "The module is deprecated and so on",
            DeprecationWarning,
        )
    """))
    assert r == [(DeprecationWarning, 'The module is deprecated and so on')]


@pytest.mark.parametrize('given, ecat, emsg', [
    ('"oh hi mark"', UserWarning, 'oh hi mark'),
    ('"oh hi mark", ImportWarning', ImportWarning, 'oh hi mark'),
    ('"oh hi mark", category=ImportWarning', ImportWarning, 'oh hi mark'),
    ('message="oh hi mark", category=ImportWarning', ImportWarning, 'oh hi mark'),
    ('category=ImportWarning, message="oh hi mark"', ImportWarning, 'oh hi mark'),

    ('"oh hi mark", garbage', Warning, 'oh hi mark'),
    ('garbage', UserWarning, 'warnings.warn(garbage)'),
    ('', UserWarning, 'warnings.warn()'),
])
def test_module_deprecated__args_parsing(given, ecat, emsg):
    r = e(WarningsExtractor, p(f"""
        import warnings
        warnings.warn({given})
    """))
    assert r == [(ecat, emsg)]



import pytest

from flake8_warnings._extractors import DecoratorsExtractor

from .helpers import e, p


@pytest.mark.parametrize('given, exp', [
    # https://github.com/tantale/deprecated
    ('deprecated(reason="use another function")', 'use another function'),
    ('deprecated.deprecated(reason="use another function")', 'use another function'),
    # https://github.com/briancurtin/deprecation
    (
        'deprecation.deprecated(details="Use the bar function instead")',
        'Use the bar function instead',
    ),
    (
        'deprecation.deprecated(deprecated_in="1.0", current_version=__version__, details="Use the bar function instead")',  # noqa
        'Use the bar function instead',
    ),
    (
        'deprecated(deprecated_in="1.0", current_version=__version__, details="Use the bar function instead")',  # noqa
        'Use the bar function instead',
    ),
    # https://github.com/mfalesni/python-deprecate
    ('deprecated(message="call to depr func")', 'call to depr func'),
    ('deprecate.deprecated(message="call to depr func")', 'call to depr func'),
    # https://github.com/multi-vac/py_deprecate
    (
        'deprecated(allowed_deprecations=[allowed_sum_caller], message="sum is no longer supported.")',  # noqa
        'sum is no longer supported.',
    ),
])
def test_extract_deprecated(given, exp):
    r = e(DecoratorsExtractor, p(f"""
        from deprecated import deprecated

        @{given}
        def some_old_function(x, y):
            return x + y
    """).body[-1])
    assert r == [(DeprecationWarning, exp)]



from typing import Iterator

import astroid

from ._base import Extractor, WarningInfo


class DocstringsExtractor(Extractor):
    """Extractor for warning messages in docstrings.
    """

    def extract(self, node: astroid.NodeNG) -> Iterator[WarningInfo]:
        yield from ()



"""Linter (flake8, pylint, custom CLI) for finding usage of deprecated functions.
"""

from ._flake8_plugin import Flake8Checker
from ._pylint_plugin import register


__version__ = '0.4.0'
__all__ = ['Flake8Checker', 'register']



from textwrap import dedent

import astroid
import pytest

from flake8_warnings._finder import WarningFinder


def p(text):
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    return tree


def e(node):
    return [(w.category, w.message) for w in WarningFinder(node).find()]


ASTROID_WARNING = "The 'astroid.node_classes' module is deprecated and will be replaced by 'astroid.nodes' in astroid 3.0.0"  # noqa


@pytest.mark.parametrize("given, etype, emsg", [
    ("import astroid.node_classes", DeprecationWarning, ASTROID_WARNING),
    ("from astroid.node_classes import *", DeprecationWarning, ASTROID_WARNING),
    ("import tests.samples.warnings_module", DeprecationWarning, "mod warn"),
    ("from tests.samples.warnings_module import fun", DeprecationWarning, "mod warn"),
    ("from tests.samples.warnings_function import func", DeprecationWarning, "func warn"),
])
def test_finder__import(given, etype, emsg):
    r = e(p(given))
    assert r == [(etype, emsg)]



from typing import Tuple, Type

from ._base import CODES, Extractor, WarningInfo
from ._decorators import DecoratorsExtractor
from ._docstrings import DocstringsExtractor
from ._stdlib import StdlibExtractor
from ._warnings import WarningsExtractor


__all__ = ["CODES", "EXTRACTORS", "Extractor", "WarningInfo"]
EXTRACTORS: Tuple[Type[Extractor], ...] = (
    DecoratorsExtractor,
    DocstringsExtractor,
    StdlibExtractor,
    WarningsExtractor,
)



import ast
import tokenize
from pathlib import Path
from typing import Iterator, Optional, Sequence

from ._finder import WarningFinder


TEMPLATE = 'WS0{code:02} {message}'


class Flake8Checker:
    name = __package__
    version = '0.0.1'

    def __init__(
        self,
        tree: ast.AST,
        file_tokens: Sequence[tokenize.TokenInfo],
        filename: Optional[str] = None,
    ) -> None:
        self._tokens = file_tokens
        self._filename = filename

    @property
    def _finder(self) -> WarningFinder:
        if not self._filename or self._filename in ('stdout', '-'):
            text = tokenize.untokenize(self._tokens).encode()
            return WarningFinder.from_text(text)
        return WarningFinder.from_path(Path(self._filename))

    def run(self) -> Iterator[tuple]:
        for winfo in self._finder.find():
            text = TEMPLATE.format(code=winfo.code, message=winfo.message)
            yield (
                winfo.line, winfo.col, text, type(self),
            )



from __future__ import annotations

from collections import deque
from contextlib import suppress
from pathlib import Path
from typing import Iterator

import astroid

from ._extractors import EXTRACTORS, Extractor, WarningInfo


class WarningFinder:
    _module: astroid.Module
    _extractors: tuple[Extractor, ...]

    def __init__(self, module: astroid.Module):
        self._module = module
        self._extractors = tuple(e() for e in EXTRACTORS)

    @classmethod
    def from_path(cls, path: Path) -> 'WarningFinder':
        text = path.read_text()
        module = astroid.parse(code=text, path=str(path))
        return cls(module)

    @classmethod
    def from_text(cls, text: str) -> 'WarningFinder':
        module = astroid.parse(code=text)
        return cls(module)

    def find(self) -> Iterator[WarningInfo]:
        yield from self._check_imports()

    def _check_imports(self) -> Iterator[WarningInfo]:
        for node in self._traverse(self._module):
            for target_node in self._get_imported_nodes(node):
                for extractor in self._extractors:
                    warnings = list(extractor.extract(target_node))
                    for warning in warnings:
                        yield warning.evolve(
                            line=node.lineno,
                            col=node.col_offset,
                            node=node,
                        )
                    # If one extractor found something for the node,
                    # don't try other extractors.
                    if warnings:
                        break

    def _get_imported_nodes(self, node) -> Iterator[astroid.NodeNG]:
        if isinstance(node, astroid.Import):
            for name, _ in node.names:
                with suppress(astroid.AstroidImportError):
                    yield self._module.import_module(name)

        if not isinstance(node, astroid.ImportFrom):
            return
        try:
            module = self._module.import_module(node.modname)
        except astroid.AstroidImportError:
            return
        yield module
        for name, _ in node.names:
            _, resolved_nodes = module.lookup(name)
            for node in resolved_nodes:
                if isinstance(node, (astroid.ClassDef, astroid.FunctionDef)):
                    yield node

    @staticmethod
    def _traverse(node: astroid.NodeNG) -> Iterator[astroid.NodeNG]:
        todo = deque([node])
        while todo:
            node = todo.popleft()
            todo.extend(node.get_children())
            yield node



from typing import Iterator, Optional

import astroid

from ._base import Extractor, WarningInfo


class DecoratorsExtractor(Extractor):
    """Extractor for deprecation decorators.

    A few examples:
        + https://github.com/tantale/deprecated
            + @deprecated
        + https://github.com/briancurtin/deprecation
            + @deprecation.deprecated
    """

    def extract(self, node: astroid.NodeNG) -> Iterator[WarningInfo]:
        if not isinstance(node, (astroid.FunctionDef, astroid.ClassDef)):
            return
        if not node.decorators:
            return
        assert isinstance(node.decorators, astroid.Decorators)
        dec: astroid.NodeNG
        for dec in node.decorators.nodes:
            warn = self._get_warning(dec)
            if warn is not None:
                yield warn

    def _get_warning(self, node: astroid.NodeNG) -> Optional[WarningInfo]:
        if not isinstance(node, astroid.Call):
            return None
        if 'deprecat' not in node.func.as_string():
            return None
        return WarningInfo(
            message=self._get_message(node),
            category=DeprecationWarning,
        )

    def _get_message(self, node: astroid.Call) -> str:
        for arg in node.args:
            msg = self._get_message_from_node(arg)
            if msg:
                return msg
        for kwarg in node.keywords:
            msg = self._get_message_from_node(kwarg.value)
            if msg:
                return msg
        return ' '.join(node.as_string().split())

    @staticmethod
    def _get_message_from_node(node: astroid.NodeNG) -> Optional[str]:
        if not isinstance(node, astroid.Const):
            return None
        if not isinstance(node.value, str):
            return None
        if ' ' not in node.value:
            return None
        return node.value



from types import MappingProxyType
from typing import Iterator, Mapping, NamedTuple, Optional, Type

import astroid


# https://docs.python.org/3/library/warnings.html
CODES: Mapping[Type[Warning], int] = MappingProxyType({
    Warning: 1,
    UserWarning: 2,
    DeprecationWarning: 3,
    SyntaxWarning: 4,
    RuntimeWarning: 5,
    FutureWarning: 6,
    PendingDeprecationWarning: 7,
    ImportWarning: 8,
    UnicodeWarning: 9,
    BytesWarning: 10,
    ResourceWarning: 11,
})
NAMES = MappingProxyType({cat.__name__: cat for cat in CODES})


class WarningInfo(NamedTuple):
    message: str
    category: Type[Warning]
    argument: Optional[str] = None
    node: Optional[astroid.NodeNG] = None
    line: int = 1
    col: int = 0

    @property
    def code(self) -> int:
        return CODES.get(self.category, 1)

    def evolve(self, **kwargs) -> 'WarningInfo':
        return self._replace(**kwargs)

    def __str__(self) -> str:
        return f'{self.line}:{self.col} [{self.category.__name__}] {self.message}'


class Extractor:
    def extract(self, node: astroid.NodeNG) -> Iterator[WarningInfo]:
        raise NotImplementedError



import pytest

from flake8_warnings._finder import WarningFinder

from .helpers import p


@pytest.mark.parametrize('given', [
    # PEP 594: Removing dead batteries
    'import aifc',
    'import asynchat',
    'import asyncore',
    'import audioop',
    'import cgi',
    'import cgitb',
    'import chunk',
    'import crypt',
    'import imghdr',
    'import mailcap',
    # 'import msilib',  # available only on windows
    'import nntplib',
    'import nis',
    'import ossaudiodev',
    'import pipes',
    'import smtpd',
    'import sndhdr',
    'import spwd',
    'import sunau',
    'import telnetlib',
    'import uu',
    'import xdrlib',

    'import distutils',
    'from distutils.core import setup',
    'from distutils.sysconfig import get_python_inc',

    'import optparse',
    'import tkinter.tix',
    'import xml.etree.cElementTree',

    # from-import
    'from smtpd import SMTPServer',
])
def test_module_deprecated(given):
    finder = WarningFinder(p(given))
    r = [(w.category, w.message) for w in finder.find()]
    assert len(r) == 1
    assert r[0][0] is DeprecationWarning



from __future__ import annotations

from textwrap import dedent

import astroid

from flake8_warnings._extractors import Extractor


def p(text) -> astroid.Module:
    """Parse the text into astroid tree, print it.
    """
    tree = astroid.parse(dedent(text))
    print(tree.repr_tree())
    return tree


def e(extractor: type[Extractor], node: astroid.NodeNG) -> list[tuple[type, str]]:
    return [(w.category, w.message) for w in extractor().extract(node)]



from typing import TYPE_CHECKING

import astroid

from ._extractors import CODES
from ._finder import WarningFinder


if TYPE_CHECKING:
    from pylint.lint import PyLinter

try:
    from pylint.checkers import BaseChecker
    from pylint.interfaces import IAstroidChecker
except ImportError:
    BaseChecker = object
    IAstroidChecker = object


CODE = 'W99{:02}'


def register(linter: "PyLinter") -> None:
    linter.register_checker(PyLintChecker(linter))


class PyLintChecker(BaseChecker):
    """
    https://pylint.pycqa.org/en/latest/how_tos/custom_checkers.html
    """

    __implements__ = IAstroidChecker

    name = 'flake8_warnings'
    msgs = {CODE.format(code): ('%s', cat.__name__, '') for cat, code in CODES.items()}

    def visit_module(self: BaseChecker, node: astroid.Module) -> None:
        finder = WarningFinder(node)
        for winfo in finder.find():
            self.add_message(
                CODE.format(winfo.code),
                line=winfo.line,
                col_offset=winfo.col,
                args=(winfo.message,),
                node=winfo.node,
            )



import warnings


def func():
    warnings.warn("func warn", DeprecationWarning)
    return 1
    warnings.warn("this one is ignored")  # type: ignore


def not_imported_func():
    warnings.warn("this one is ignored")



import sys
from argparse import ArgumentParser
from pathlib import Path
from typing import Iterator, List, NoReturn, TextIO

from ._finder import WarningFinder


def get_paths(path: Path) -> Iterator[Path]:
    """Recursively yields python files.
    """
    if not path.exists():
        raise FileNotFoundError(str(path))
    if path.is_file():
        if path.suffix == '.py':
            yield path
        return
    for subpath in path.iterdir():
        if subpath.name[0] == '.':
            continue
        if subpath.name == '__pycache__':
            continue
        yield from get_paths(subpath)


def main(argv: List[str], stream: TextIO) -> int:
    parser = ArgumentParser()
    parser.add_argument('paths', nargs='+')
    args = parser.parse_args(argv)
    found_warnings = 0
    for root in args.paths:
        for path in get_paths(Path(root)):
            path_shown = False
            finder = WarningFinder.from_path(path)
            for warning in finder.find():
                if not path_shown:
                    print(path, file=stream)
                    path_shown = True
                print(' ', warning, file=stream)
                found_warnings += 1
    return min(found_warnings, 100)


def entrypoint() -> NoReturn:
    sys.exit(main(sys.argv[1:], sys.stdout))



from typing import Iterator, List, Optional

import astroid

from ._base import NAMES, Extractor, WarningInfo


BRANCHING = (
    astroid.If,
    astroid.With,
    astroid.TryExcept,
    astroid.TryFinally,
    astroid.Return,
    astroid.Raise,
)


class WarningsExtractor(Extractor):
    """Extractor for `wanings.warn()` invocations.
    """

    def extract(self, node: astroid.NodeNG) -> Iterator[WarningInfo]:
        if isinstance(node, (astroid.FunctionDef, astroid.Module)):
            yield from self._extract_from_body(node.body)

    def _extract_from_body(self, body: List[astroid.NodeNG]) -> Iterator[WarningInfo]:
        for node in body:
            if isinstance(node, astroid.Expr):
                node = node.value
            if isinstance(node, BRANCHING):
                return
            warning = self._get_warning(node)
            if warning is not None:
                yield warning

    def _get_warning(self, node: astroid.NodeNG) -> Optional[WarningInfo]:
        # check if it is a call to `warnings.warn`
        if not isinstance(node, astroid.Call):
            return None
        if node.func.as_string() != 'warnings.warn':
            return None
        return WarningInfo(
            message=self._get_message(node),
            category=NAMES.get(self._get_category(node), Warning),
        )

    @staticmethod
    def _get_message(node: astroid.Call) -> str:
        # extract positional category
        if node.args:
            arg_node = node.args[0]
            if isinstance(arg_node, astroid.Const):
                return str(arg_node.value)

        # extract keyword category
        for kwarg in node.keywords:
            if kwarg.arg != 'message':
                continue
            if isinstance(kwarg.value, astroid.Const):
                return str(kwarg.value.value)
        return ' '.join(node.as_string().split())

    @staticmethod
    def _get_category(node: astroid.Call) -> str:
        # extract positional category
        if len(node.args) > 1:
            arg_node = node.args[1]
            if isinstance(arg_node, astroid.Name):
                return arg_node.name

        # extract keyword category
        for kwarg in node.keywords:
            if kwarg.arg != 'category':
                continue
            arg_node = kwarg.value
            if isinstance(arg_node, astroid.Name):
                return arg_node.name
        return 'UserWarning'



from typing import Iterator

import astroid

from ._base import Extractor, WarningInfo


MODULES = frozenset({
    # PEP 594: Removing dead batteries
    'aifc',
    'asynchat',
    'asyncore',
    'audioop',
    'cgi',
    'cgitb',
    'chunk',
    'crypt',
    'imghdr',
    'mailcap',
    'msilib',
    'nntplib',
    'nis',
    'ossaudiodev',
    'pipes',
    'smtpd',
    'sndhdr',
    'spwd',
    'sunau',
    'telnetlib',
    'uu',
    'xdrlib',

    # PEP 632, removed in Python 3.12
    'distutils',

    # deprecated but not announced to be removed
    'optparse',
    'tkinter.tix',
    'xml.etree.cElementTree',
})


class StdlibExtractor(Extractor):
    """Extractor for warning messages in docstrings.
    """

    def extract(self, node: astroid.NodeNG) -> Iterator[WarningInfo]:
        if not isinstance(node, astroid.Module):
            return
        qname = node.qname()
        if qname not in MODULES:
            qname = qname.split('.')[0]
        if qname in MODULES:
            yield WarningInfo(
                message=f'stdlib module {qname} is deprecated',
                category=DeprecationWarning,
            )



from flake8_warnings._extractors._base import CODES, NAMES


def test_codes():
    assert len(CODES) == len(set(CODES.values()))
    last_code = max(CODES.values())
    assert set(CODES.values()) == set(range(1, last_code + 1))


def test_names():
    assert NAMES['UserWarning'] is UserWarning
